from sandman.model import register, Model

class Artist(Model):
    __tablename__ = 'Artist'


class Track(Model):
    __tablename__ = 'Track'

    @staticmethod
    def validate_PUT(resource=None):
        if int(resource.TrackId) == 999:
            return False
        return True

class Album(Model):
    __tablename__ = 'Album'
    __methods__ = ('POST', 'PATCH', 'DELETE', 'PUT', 'GET')

class Playlist(Model):
    __tablename__ = 'Playlist'
    __methods__ = ('POST', 'PATCH')

class Genre(Model):
    __tablename__ = 'Genre'
    __endpoint__ = 'styles'
    __methods__ = ('GET', 'DELETE')

    @staticmethod
    def validate_GET(resource=None):
        if isinstance(resource, list):
            return True
        elif resource and resource.GenreId == 1:
            return False
        return True

register((Artist, Album, Playlist, Track))
register(Genre)



"""The model module is repsonsible exposes the :class:`sandman.Resource` class, from which user
models should derive. It also makes the :func:`register` function available, which
maps endpoints to their associated classes."""

from decimal import Decimal
from . import db, app
from sqlalchemy.ext.declarative import declarative_base, DeferredReflection
from flask import current_app

__all__ = ['Model', 'register']

def register(cls):
    """Register with the API a :class:`sandman.Resource` class and associated
    endpoint.
    
    :param cls: User-defined class derived from :class:`sandman.Resource` to be
                registered with the endpoint returned by :func:`endpoint()`
    :type cls: :class:`sandman.Resource` or tuple
    
    """
    with app.app_context():
        if getattr(current_app, 'endpoint_classes', None) is None:
            current_app.endpoint_classes = {}
        if isinstance(cls, (list, tuple)):
            for entry in cls:
                current_app.endpoint_classes[entry.endpoint()] = entry
        else:
            current_app.endpoint_classes[cls.endpoint()] = cls
    Model.prepare(db.engine)

       
class Resource(object):
    """A mixin class containing the majority of the RESTful API functionality.
    
    :class:`sandman.Resource` is the base class of `:class:`sandman.Model`,
    from which user models are derived.
    """

    # override :attr:`__endpoint__` if you wish to configure the
    # :class:`sandman.Resource`'s endpoint. 
    #
    # Default: __tablename__ in lowercase and pluralized
    __endpoint__ = None

    # override :attr:`__methods__` if you wish to change the HTTP methods 
    # this :class:`sandman.Resource` supports.
    #
    # Default: ``('GET', 'POST', 'PATCH', 'DELETE', 'PUT')``
    __methods__ = ('GET', 'POST', 'PATCH', 'DELETE', 'PUT')

    @classmethod
    def endpoint(cls):
        """Return the :class:`sandman.Resource`'s endpoint.

        :rtype: string

        """
        if cls.__endpoint__ is not None:
            return cls.__endpoint__
        return cls.__tablename__.lower() + 's'

    def resource_uri(self):
        """Return the URI at which the resource can be found.
        
        :rtype: string

        """ 
        primary_key_value = getattr(self, self.primary_key(), None)
        return '/{}/{}'.format(self.endpoint(), primary_key_value)

    def links(self):
        """Return a list of links for endpoints related to the resource."""
        links = []
        links.append({'rel': 'self', 'uri': self.resource_uri()})
        return links

    @classmethod
    def primary_key(cls):
        """Return the name of the table's primary key
        
        :rtype: string

        """

        return cls.__table__.primary_key.columns.values()[0].name

    def as_dict(self):
        """Return a dictionary containing only the attributes which map to 
        an instance's database columns.
        
        :rtype: dict

        """ 
        result_dict = {}
        for column in self.__table__.columns.keys():
            result_dict[column] = getattr(self, column, None)
            if isinstance(result_dict[column], Decimal):
                result_dict[column] = str(result_dict[column])
        result_dict['links'] = self.links()
        return result_dict

    def from_dict(self, dictionary):
        """Set a set of attributes which correspond to the 
        :class:`sandman.Resource`'s columns.

        :param dict dictionary: A dictionary of attributes to set on the
        instance whose keys are the column names of the :class:`sandman.Resource`'s
        underlying database table. 
        
        """ 
        for column in self.__table__.columns.keys():
            value = dictionary.get(column, None)
            if value:
                setattr(self, column, value)

    def replace(self, dictionary):
        """Set all attributes which correspond to the 
        :class:`sandman.Resource`'s columns to the values in *dictionary*,
        inserting None if an attribute's value is not specified.

        :param dict dictionary: A dictionary of attributes to set on the
        instance whose keys are the column names of the :class:`sandman.Resource`'s
        underlying database table. 

        """
        for column in self.__table__.columns.keys():
                setattr(self, column, None)
        self.from_dict(dictionary)
 
Model = declarative_base(cls=(DeferredReflection, Resource))



"""Main test class for sandman"""

from sandman import app, db
import unittest
import os
import shutil
import json

class TestSandman:
    DB_LOCATION = os.path.join(os.getcwd(), 'sandman', 'test', 'chinook')

    def setup_method(self, method):
        shutil.copy(os.path.join(os.getcwd(), 'sandman', 'test', 'data', 'chinook'), self.DB_LOCATION) 
        app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:////' + self.DB_LOCATION
        app.config['TESTING'] = True
        self.app = app.test_client()
        from . import models

    def teardown_method(self, method):
        os.unlink(self.DB_LOCATION)
        self.app = None

    def get_response(self, uri, status_code, has_data=True, headers=None):
        if headers is None:
            headers = {}
        response = self.app.get(uri, headers=headers)
        assert response.status_code == status_code
        if has_data:
            assert response.data
        return response

    def post_response(self):
        response = self.app.post('/artists', 
                content_type='application/json', 
                data=json.dumps({u'Name': u'Jeff Knupp'}))
        assert response.status_code == 201
        assert json.loads(response.data)[u'Name'] == u'Jeff Knupp'
        return response

    @staticmethod
    def is_html_response(response):
        return '<!DOCTYPE html>' in response.data

    def test_get(self):
        response = self.get_response('/artists', 200)
        assert len(json.loads(response.data)[u'resources']) == 275

    def test_post(self):
        response = self.post_response()
        assert json.loads(response.data)[u'Name'] == u'Jeff Knupp'
        assert json.loads(response.data)[u'links'] == [{u'rel': u'self', u'uri': u'/artists/276'}]

    def test_posted_location(self):
        post_response = self.post_response()
        location = post_response.headers['Location']
        response = self.get_response(location, 200)

    def test_posted_uri(self):
        post_response = self.post_response()
        as_json = json.loads(post_response.data)
        uri = as_json[u'links'][0][u'uri']
        response = self.app.get(uri)
        assert as_json[u'Name'] == u'Jeff Knupp'

    def test_patch_new_resource(self):
        response = self.app.patch('/artists/276', 
                content_type='application/json', 
                data=json.dumps({u'Name': u'Jeff Knupp'}))
        assert response.status_code == 201
        assert type(response.data) == str
        assert json.loads(response.data)['Name'] == u'Jeff Knupp'
        assert json.loads(response.data)['links'] == [{u'rel': u'self', u'uri': u'/artists/276'}]

    def test_patch_existing_resource(self):
        response = self.app.patch('/artists/275', 
                content_type='application/json', 
                data=json.dumps({u'Name': u'Jeff Knupp'}))
        assert response.status_code == 204
        response = self.get_response('/artists/275', 200)
        assert json.loads(response.data.decode('utf-8'))[u'Name'] == u'Jeff Knupp'
        assert json.loads(response.data.decode('utf-8'))[u'ArtistId'] == 275

    def test_delete_resource(self):
        response = self.app.delete('/artists/275')
        assert response.status_code == 204
        response = self.get_response('/artists/275', 404, False)

    def test_delete_non_existant_resource(self):
        response = self.app.delete('/artists/404')
        assert response.status_code == 404

    def test_delete_not_supported(self):
        response = self.app.delete('/playlists/1')
        assert response.status_code == 403

    def test_unsupported_resource_method(self):
        response = self.app.patch('/styles/26',
                content_type='application/json', 
                data=json.dumps({u'Name': u'Hip-Hop'}))
        assert response.status_code == 403

    def test_unsupported_collection_method(self):
        response = self.app.get('/playlists')
        assert response.status_code == 403

    def test_user_defined_endpoint(self):
        response = self.get_response('/styles', 200)
        assert len(json.loads(response.data)[u'resources']) == 25

    def test_user_validation(self):
        self.get_response('/styles/1', 403, False)

    def test_get_html(self):
        response = self.get_response('/artists/1', 200, headers={'Accept': 'text/html'})
        assert self.is_html_response(response)

    def test_get_html_collection(self):
        response = self.app.get('/artists', 200, headers={'Accept': 'text/html'})
        assert self.is_html_response(response)
        assert 'Aerosmith' in response.data

    def test_get_json(self):
        response = self.get_response('/artists', 200, headers={'Accept': 'application/json'})
        assert len(json.loads(response.data)[u'resources']) == 275

    def test_post_html_response(self):
        response = self.app.post('/artists', 
                content_type='application/json', 
                headers={'Accept': 'text/html'},
                data=json.dumps({u'Name': u'Jeff Knupp'}))
        assert response.status_code == 201
        assert 'Jeff Knupp' in response.data

    def test_put_resource(self):
        response = self.app.put('/tracks/1', 
                content_type='application/json', 
                data=json.dumps(
                    {'Name': 'Some New Album',
                      'AlbumId': 1,
                      'GenreId': 1,
                      'MediaTypeId': 1,
                      'Milliseconds': 343719,
                      'TrackId': 1,
                      'UnitPrice': 0.99,}))
        assert response.status_code == 204
        response = self.get_response('/tracks/1', 200)
        assert json.loads(response.data.decode('utf-8'))[u'Name'] == u'Some New Album'
        assert json.loads(response.data.decode('utf-8'))[u'Composer'] is None

    def test_put_unknown_resource(self):
        response = self.app.put('/tracks/99999', 
                content_type='application/json', 
                data=json.dumps(
                    {'Name': 'Some New Album',
                      'AlbumId': 1,
                      'GenreId': 1,
                      'MediaTypeId': 1,
                      'Milliseconds': 343719,
                      'TrackId': 99999,
                      'UnitPrice': 0.99,}))
        assert response.status_code == 404


    def test_put_invalid_foreign_key(self):
        response = self.app.put('/tracks/998', 
                content_type='application/json', 
                data=json.dumps(
                    {'Name': 'Some New Album',
                      'Milliseconds': 343719,
                      'TrackId': 998,
                      'UnitPrice': 0.99,}))
        assert response.status_code == 422

    def test_put_fail_validation(self):
        response = self.app.put('/tracks/999', 
                content_type='application/json', 
                data=json.dumps(
                    {'Name': 'Some New Album',
                      'GenreId': 1,
                      'AlbumId': 1,
                      'MediaTypeId': 1,
                      'Milliseconds': 343719,
                      'TrackId': 999,
                      'UnitPrice': 0.99,}))
        assert response.status_code == 403



"""
sandman
-------

**sandman** "makes things REST". Have an existing database you'd like to expose via
a REST API? Normally, you'd have to write a ton of boilerplate code for
the ORM you're using. 

We're programmers. We don't write boilerplate.

Simple Setup
````````````

.. code:: python

    from sandman import app, db

    app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///chinook'

    from sandman.model import register, Model

    class Artist(Model):
        __tablename__ = 'Artist'

    class Album(Model):
        __tablename__ = 'Album'

    class Playlist(Model):
        __tablename__ = 'Playlist'

    register((Artist, Album, Playlist))

    app.run()

Let's start our new API server and make a request.

.. code:: bash

    $ python runserver.py &
    * Running on http://127.0.0.1:5000/

    $ curl GET http://localhost:5000/artists

Here is the JSON returned:

.. code:: json

    {
        "ArtistId": 273,
        "Name": "C. Monteverdi, Nigel Rogers - Chiaroscuro; London Baroque; London Cornett & Sackbu",
        "links": [
        {
            "rel": "self",
            "uri": "/artists/ArtistId"
        }
        ]
    },
    {
        "ArtistId": 274,
        "Name": "Nash Ensemble",
        "links": [
        {
            "rel": "self",
            "uri": "/artists/ArtistId"
        }
        ]
    },
    {
        "ArtistId": 275,
        "Name": "Philip Glass Ensemble",
        "links": [
        {
            "rel": "self",
            "uri": "/artists/ArtistId"
        }
        ]
    }
    ]

Batteries Included
``````````````````

With **sandman**, (almost) zero boilerplate code is required. Your existing database
structure and schema is introspected and your database tables magically get a
RESTful API. For each table, Sandman creates:

- proper endpoints 
 
- support for a configurable set of HTTP verbs 
     
    - GET

    - POST

    - PATCH

    - DELETE

- responses with appropriate ``rel`` links automatically

- essentially a HATEOAS-based service sitting in front of your database

*Warning: Sandman is still very much a work in progress.* It is not suitable for
use **anywhere.** Don't use it for anything important. It's also often changing 
in backwards incompatible ways.

Links
`````

* `website <http://www.github.com/jeffknupp/sandman/>`_
* `documentation <http://pythonhosted.org/sandman/>`_

"""

from __future__ import print_function
from setuptools import setup
from setuptools.command.test import test as TestCommand
import sys

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

setup(
    name='sandman',
    version='0.2.3.4',
    url='http://github.com/jeffknupp/sandman/',
    license='Apache Software License',
    author='Jeff Knupp',
    tests_require=['pytest'],
    install_requires=['Flask>=0.10.1',
                      'Flask-SQLAlchemy>=1.0',
                      'SQLAlchemy==0.8.2',
                      ],
    cmdclass={'test': PyTest},
    author_email='jeff@jeffknupp.com',
    description='Automated REST APIs for existing database-driven systems',
    long_description=__doc__,
    packages=['sandman', 'sandman.test'],
    include_package_data=True,
    platforms='any',
    test_suite='sandman.test.test_sandman',
    classifiers = [
        'Programming Language :: Python',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        ],
)



"""Sandman REST API creator for Flask and SQLAlchemy"""

from flask import (jsonify, request, g, 
        current_app, Response, render_template,
        make_response)
from sqlalchemy.exc import IntegrityError
from . import app, db
from .exception import JSONException

JSON, HTML = range(2)

def _get_session():
    """Return (and memoize) a database session"""
    session = getattr(g, '_session', None)
    if session is None:
        session = g._session = db.session()
    return session

def _get_mimetype(current_request):
    """Return the mimetype for this request."""
    if 'Accept' not in current_request.headers:
        return JSON

    if 'json' in current_request.headers['Accept']:
        return JSON
    else:
        return HTML

def _single_resource_json_response(resource):
    """Return the JSON representation of *resource*"""
    return jsonify(**resource.as_dict())
   
def _single_resource_html_response(resource):
    """Return the HTML representation of *resource*"""
    tablename = resource.__tablename__
    resource.pk = getattr(resource, resource.primary_key())
    resource.attributes = resource.as_dict()
    return make_response(render_template('resource.html', resource=resource,
        tablename=tablename))

def _collection_json_response(resources):
    """Return the HTML representation of the collection *resources*"""
    result_list = []
    for resource in resources:
        result_list.append(resource.as_dict())
    return jsonify(resources=result_list)

def _collection_html_response(resources):
    """Return the HTML representation of the collection *resources*"""
    return make_response(render_template('collection.html',
        resources=resources))

def _validate(cls, method, resource=None):
    """Return ``True`` if the the given *cls* supports the HTTP *method* found 
    on the incoming HTTP request.
    
    :param cls: class associated with the request's endpoint
    :type cls: :class:`sandman.model.Resource` instance
    :param string method: HTTP method of incoming request
    :param resource: *cls* instance associated with the request
    :type resource: :class:`sandman.model.Resource` or None
    :rtype: bool

    """
    if not method in cls.__methods__:
        return False

    class_validator_name = 'validate_' + method

    if hasattr(cls, class_validator_name):
        class_validator = getattr(cls, class_validator_name)
        return class_validator(resource)

    return True

def endpoint_class(collection):
    """Return the :class:`sandman.model.Resource` associated with the endpoint
    *collection*.

    :param string collection: a :class:`sandman.model.Resource` endpoint
    :rtype: :class:`sandman.model.Resource`

    """
    session = _get_session()
    with app.app_context():
        return current_app.endpoint_classes[collection]

def retrieve_resource(collection, key):
    """Return the resource of type *cls* identified by key *key*.

    :param string collection: a :class:`sandman.model.Resource` endpoint
    :param string key: primary key of resource
    :rtype: class:`sandman.model.Resource`
    """
    session = _get_session()
    with app.app_context():
        cls = current_app.endpoint_classes[collection]
    return session.query(cls).get(key)

def resource_created_response(resource, current_request):
    """Return HTTP response with status code *201*, signaling a created *resource*
    
    :param resource: resource created as a result of current request
    :type resource: :class:`sandman.model.Resource`
    :rtype: :class:`flask.Response`
    """
    if _get_mimetype(current_request) == JSON:
        response = _single_resource_json_response(resource)
    else:
        response = _single_resource_html_response(resource)
    response.status_code = 201
    response.headers['Location']  = 'http://localhost:5000/' + resource.resource_uri()
    return response

def resource_response(resource, current_request):
    result_dict = resource.as_dict()
    if _get_mimetype(current_request) == JSON:
        return _single_resource_json_response(resource)
    else:
        return _single_resource_html_response(resource)


def unsupported_method_response():
    """Return the appropriate *Response* with status code *403*, signaling the HTTP method used
    in the request is not supported for the given endpoint.
    
    :rtype: :class:`flask.Response`

    """
    response = Response()
    response.status_code = 403
    return response

def no_content_response():
    """Return the appropriate *Response* with status code *204*, signaling a completed action
    which does not require data in the response body
    
    :rtype: :class:`flask.Response`
    """
    response = Response()
    response.status_code = 204
    return response

def update_resource(resource, data):
    """Replace the contents of a resource with *data* and return an appropriate
    *Response*.
    
    :param resource: :class:`sandman.model.Resource` to be updated
    :param data: New values for the fields in *resource*

    """

    resource.from_dict(request.json)
    session = _get_session()
    session.merge(resource)
    session.commit()
    return no_content_response()


@app.route('/<collection>/<key>', methods=['PATCH'])
def patch_resource(collection, key):
    """"Upsert" a resource identified by the given key and return the appropriate
    *Response*.
    
    If no resource currently exists at `/<collection>/<key>`, create it
    with *key* as its primary key and return a
    :func:`resource_created_response`.

    If a resource *does* exist at `/<collection>/<key>`, update it with
    the data sent in the request and return a :func:`no_content_response`.

    Note: HTTP `PATCH` (and, thus, :func:`patch_resource`) is idempotent

    :param string collection: a :class:`sandman.model.Resource` endpoint
    :param string key: the primary key for the associated :class:`sandman.model.Resource`
    :rtype: :class:`flask.Response`

    """
    session = _get_session()
    with app.app_context():
        cls = current_app.endpoint_classes[collection]

    resource = session.query(cls).get(key)

    if not _validate(cls, request.method, resource):
        return unsupported_method_response()

    if resource is None:
        resource = cls()
        resource.from_dict(request.json)
        setattr(resource, resource.primary_key(), key)
        session.add(resource)
        session.commit()
        return resource_created_response(resource, request)
    else:
        return update_resource(resource, request.json)

@app.route('/<collection>/<key>', methods=['PUT'])
def replace_resource(collection, key):
    """Replace the resource identified by the given key and return the appropriate
    response.

    :param string collection: a :class:`sandman.model.Resource` endpoint
    :rtype: :class:`flask.Response`

    """
    resource = retrieve_resource(collection, key)

    if resource is None:
        return JSONException('Requested resource not found', code=404)
    elif not _validate(endpoint_class(collection), request.method, resource):
        return unsupported_method_response()

    resource.replace(request.json)
    session = _get_session()
    session.add(resource)
    try:
        session.commit()
    except IntegrityError as e:
        return JSONException('Requested resource not found. Exception: [{}]'.format(e.message), code=422)
    return no_content_response()

@app.route('/<collection>', methods=['POST'])
def add_resource(collection):
    """Return the appropriate *Response* based on adding a new resource to *collection*

    :param string collection: a :class:`sandman.model.Resource` endpoint
    :rtype: :class:`flask.Response`

    """
    cls = endpoint_class(collection)
    resource = cls()
    resource.from_dict(request.json)

    if not _validate(cls, request.method, resource):
        return unsupported_method_response()

    session = _get_session()
    session.add(resource)
    session.commit()
    return resource_created_response(resource, request)

@app.route('/<collection>/<key>', methods=['DELETE'])
def delete_resource(collection, key):
    """Return the appropriate *Response* for deleting an existing resource in *collection*
    
    :param string collection: a :class:`sandman.model.Resource` endpoint
    :param string key: the primary key for the associated :class:`sandman.model.Resource`
    :rtype: :class:`flask.Response`

    """
    with app.app_context():
        cls = current_app.endpoint_classes[collection]
    resource = cls()
    session = _get_session()
    resource = session.query(cls).get(key)

    if resource is None:
        return JSONException('Requested resource not found', code=404)
    elif not _validate(endpoint_class(collection), request.method, resource):
        return unsupported_method_response()

    session.delete(resource)
    session.commit()
    return no_content_response()


@app.route('/<collection>/<key>', methods=['GET'])
def show_resource(collection, key):
    """Return the appropriate *Response* for retrieving a single resource
    
    :param string collection: a :class:`sandman.model.Resource` endpoint
    :param string key: the primary key for the associated :class:`sandman.model.Resource`
    :rtype: :class:`flask.Response`

    """
    resource = retrieve_resource(collection, key)
    
    if resource is None:
        return JSONException('Requested resource not found', code=404)
    elif not _validate(endpoint_class(collection), request.method, resource):
        return unsupported_method_response()

    return resource_response(resource, request)

@app.route('/<collection>', methods=['GET'])
def show_collection(collection):
    """Return the appropriate *Response* for retrieving a collection of resources
    
    :param string collection: a :class:`sandman.model.Resource` endpoint
    :param string key: the primary key for the associated :class:`sandman.model.Resource`
    :rtype: :class:`flask.Response`

    """
    with app.app_context():
        cls = current_app.endpoint_classes[collection]
    session = _get_session()
    resources = session.query(cls).all()

    if not _validate(endpoint_class(collection), request.method, resources):
        return unsupported_method_response()

    if _get_mimetype(request) == JSON:
        return _collection_json_response(resources)
    else:
        return _collection_html_response(resources)



"""Exception specifications for Sandman"""

from werkzeug.exceptions import HTTPException
from flask import jsonify

class JSONException(HTTPException):
    """Excecption which generates a :class:`flask.Response` object whose
    *data* is JSON rather than HTML"""

    def __init__(self, description=None, code=400, response=None):
        super(JSONException, self).__init__(description, response)
        self.code = code

    def get_headers(self, environ=None):
        """Return the appropriate content-type: *application/json*"""
        return[('Content-type', 'application/json')]

    def get_body(self, environ=None):
        """Return the body of the *reponse* serialized as JSON"""
        return jsonify({'result': False, 'message': self.message})
